# ====================================================================
# Fitting SDT-MAX to simultaneous lineup data using Maximum Likelihood Estimation
# 
# By Matthew Kaesler, John Dunn and Carolyn Semmler
# December 2019
# 
# ====================================================================
#
# Example of correct data format for this model fit routine, simultaneous data from Kaesler, Dunn, Ransom & Semmler
#                           
#               Confidence Bin   100-91 90-81 80-66 65-51 50-0 Reject
#                    Criterion     c5    c4    c3    c2    c1  
# Target IDs (TID)                 24    25    30    9     11    NaN    
# TP Foil IDs + TID (TD)           24    26    35    13    22    19     
# Target Absent Foil IDs (FA)      4     11    25    16    24    61   
#
# Confidence bins/criteria are ordered most conservative (c5) to least conservative (c1)
#
# This code can be used for data with any number of criteria > 2. Single criteria datasets cannot be fit because 
# the model becomes saturated, i.e. once d and c are estimated there are no degrees of freedom
# left to estimate s. An equal variance model is required in this case.
# 
# A very large number of criteria may lead the model to become overfit and/or lead to low cell counts, which may 
# affect interpretation of the chi-square test that is used to assess model fit. 
# ====================================================================

#Set up observed data in format shown above

obsData <- matrix(data = c(24,25,30,9,11,NaN, 
                            24,26,35,13,22,19,
                            4,11,25,16,24,61),

                   nrow = 3,
                   ncol = 6,
                   byrow = TRUE)

n <- 6 # lineup size, important for the functions that predict the data, which are defined below

# =====================================================================

# Functions for predicting data and comapring predicted data to observed data

# Likelihood functions for SDT-MAX, calculate the probability of each model outcome
# Derivations of these functions can be found in Appendix A of Kaesler, Dunn, Ransom and Semmler 

# Probability of Target ID (TID) on TP trials according to MAX model
QT <- function(c,d,s,n){
  m <- function(x) dnorm(x,mean = d, sd = s)*pnorm(x)^(n-1)
  p <- rep(0,length(c))
  for (i in 1:length(c)){
    a <- integrate(m,c[i],15) 
    p[i] <- a$value
  }
  return(p)
}

# Probability of detection (TD) on TP trials MAX model
# This is any detection, both TID and TP foil ID
TP <- function(c,d,s,n){
  p <- rep(0,length(c))
  for (i in 1:length(c)){
    p[i] <- pnorm(((c[i])-d)/s)*pnorm(c[i])^(n-1)
  }
  p <- 1 - p
  return(p)
}

# Probability of TA foil ID (FA) MAX model
TA <- function(c,n){
  p <- rep(0,length(c))
  for (i in 1:length(c)){
    p[i] = pnorm(c[i])^n
  }
  p = 1 - p
  return(p)
}

# ==========================================================================

# This is the wrapper function that calls the likelihood functions and generates the data

genPred <- function(pars, obsData, n){
  
  # Unpack vector of parameters for use by likelihood functions
  c <- pars[1:(length(pars)-2)] 
  d <- pars[length(pars)-1] # second-to-last parameter is always d
  s <- tail(pars,1) # last parameter is always s
  
  # Calculate total number of TP and TA lineups
  totalTP <- sum(obsData[2,])
  totalTA <- sum(obsData[3,]) 
  
  TID <- QT(c(c, -15),d,s,n) 
  TID <- c(TID[1],diff(TID))
  
  TD <- c(TP(c,d,s,n),1) 
  TD <- c(TD[1],diff(TD))
  
  FA <- c(TA(c,n),1)
  FA <- c(FA[1],diff(FA))
  
  # Convert proportions to counts
  TID <- TID*totalTP
  TD <- TD*totalTP
  FA <- FA*totalTA
  
  predData <- rbind(TID,TD,FA)
  rownames(predData) <- c()
  
  return(predData)
}

# ==========================================================================

# Chi-square function, calls genPred and compares the observed data to the
# predicted data at each iteration, calculates the fit value accordingly. 

chiSq <- function(pars,obsData,n){
  
  predData <- genPred(pars,obsData,n)
  lastcell <- ncol(obsData)  
  ncrit <- ncol(obsData)-1 #number of criteria
  f <- rep(0,nrow(obsData)+2) #for storing and summing chi-sq fit value
  
  for (i in 1:ncrit){
    
    # TID
    a <- predData[1,i] 
    b <- obsData[1,i]
    f[1] <- f[1] + (b-a)^2/a
    
    # Foil ID on TP lineup
    a <- predData[2,i]-predData[1,i] 
    b <- obsData[2,i]-obsData[1,i] 
    f[2] = f[2] + (b-a)^2/a
    
    # FA
    a <- predData[3,i] 
    b <- obsData[3,i]
    f[3] <- f[3] + (b-a)^2/a
  }
  
  # Incorrect Rejection TP
  a <- predData[2,lastcell] 
  b <- obsData[2,lastcell]
  f[4] <- (b-a)^2/a
  
  # Correct Rejection TA
  a <- predData[3,lastcell] 
  b <- obsData[3,lastcell]
  f[5] <- (b-a)^2/a
  
  f <- sum(f)
  
  return(f)  
}
# ====================================================================

# Optimsation setup and model fit
# This section of the code sets up the optimsation and minmises the chiSq function defined above

# Define function to set starting values

startVals <- function(obsData,n) {

  ncrit <- ncol(obsData)-1 #number of criteria 
  
  # Random-like criteria starting points
  c0 <- obsData[nrow(obsData),] 
  c0[length(c0)] <- max(c(c0,0.5))
  a0 <- cumsum(c0)/sum(c0)
  c0 <- qnorm((1-a0)^(1/n))
  c0 <- c0[1:length(c0)-1] 
  
  # Concatenate to form starting parameter vector in order accepted by model functions
  x0 = c(c0,1.5,1) #cMax, ..., c1, dt, s
  
  return(x0)
}

# Call function to generate starting values

x0 <- startVals(obsData,n)


# Set inequality constraint matrix for criteria, stops them crossing over

ncrit <- ncol(obsData)-1 # number of criteria 
nparam <- length(x0) # number of parameters
A <- matrix(0,ncrit-1,nparam) # k x p matrix
b <- rep(0,ncrit-1) # vector of >= constraints to satisfy
for (i in 1:ncrit-1) {
  A[i,i] <- 1; A[i,i+1] <- -1
}

# Minimise chiSq function using the starting values and constraints defined above

out <- constrOptim(theta = x0, f = chiSq, grad = NULL, ui = A, ci = b, mu = 1e-04, method = "Nelder-Mead",
                   outer.iterations = 100, obsData = obsData, n = n)

# =========================================================================

# Extract relevant results from optimsation output

# Degrees of freedom and p-value, this is done when collating the results list below
df <- (nrow(obsData)*ncrit) - nparam
pval <- pchisq(out$value, df = df, lower.tail = FALSE)
p

# Predicted Data - Useful for plotting model ROCs
predData <- genPred(out$par, obsData, n)
predData[1,ncol(predData)] <- NaN #replace meaningless number in reject cell of pred Data with NaN

# Collate in to list for easy viewing
results <- list(
  c = out$par[1:(length(out$par)-2)], # same as extracting pars in genpred
  d = out$par[length(out$par)-1], 
  s = tail(out$par,1), 
  x0 = x0, 
  n = n, 
  nparam = nparam, 
  fitVal = out$value, 
  df = df, 
  pval = pval, 
  obs = obsData, 
  pred = predData, 
  converge = out$convergence)

# ==========================================================================

